'''
Parsing arguments and input data
'''
# standard command-line argument parser
import argparse
# utilities for working with paths
import os
from os import listdir
from os.path import isfile, join, basename
# system utilities
import sys


# default input file for the case of no '--in' specified
default_in = "data.txt"

# setup command line arguments
parser = argparse.ArgumentParser(description='Solve system with enchanced simplex method.')
parser.add_argument('--in', type=str, help='input file or directory, "%s" by default' % (default_in, ), metavar='input', default=default_in, dest='in_file')
parser.add_argument('--out', type=str, help='output file or directory. System stdout by default', metavar='output', dest='out_file')
parser.add_argument('--method', type=str, choices=['2-phase', '1-phase', 'gauss'], default='2-phase')

# parse command line arguments
args = parser.parse_args()

class Controller:
    '''
    Main controller, analyses command-line arguments, detects output and input files. 
    Runs corresponding function for file processing and problem solving.
    '''
    def __init__(self, args):
        '''
        Object constructor
            :param args: parsed command-line arguments
        '''
        self.__method = args.method
        # checks given input file or directory exists
        if not os.path.exists(args.in_file):
             sys.exit('%s doesn\' exists' % (args.in_file, ))
        if args.out_file is not None:
            # checks output file conforms input: for input file - no output or output regular file, for input directory - no output or output directory. 
            if os.path.isfile(args.in_file):
                if os.path.exists(args.out_file) and os.path.isdir(args.out_file):        
                    sys.exit('%s is file, but %s is directory' % (args.in_file, args.out_file, ))
            elif os.path.isdir(args.in_file):
                if os.path.exists(args.out_file) and os.path.isfile(args.out_file):        
                    sys.exit('%s is directory, but %s is file' % (args.in_file, args.out_file, ))

        self.__single_file = os.path.isfile(args.in_file)
 
        # creates set of tasks with given input and corresponding output
        if self.__single_file:
            # creates single task with regular input file and given output
            self.__files = [(args.in_file, args.out_file)]
        else:
            # Creates set of task with each file in directory and and corresponding output. In the case output is directory then 
            # each file '<dir_in>'/'<file>.txt' will produce '<dir_out>'/'<file>_out.txt'
            if args.out_file is None:
                self.__files = [(join(args.in_file, f), None) for f in listdir(args.in_file) if isfile(join(args.in_file, f))]
            else:
                if not os.path.exists(args.out_file):
                    os.makedirs(args.out_file)
                self.__files = [(join(args.in_file, f), join(args.out_file, os.path.splitext(f)[0] + "_out.txt")) for f in listdir(args.in_file) if isfile(join(args.in_file, f))]
        
    def get_method(self):
        '''
        Returns user-defined run method
        '''
        return self.__method

    def run(self, run_function, read_data_function):
        '''
        Runs given function task on the set of previously generated tasks
            :param run_function: function to be applied to given task. Should accept exactly 2 arguments: input file name and output stream.
        '''
        for t in self.__files:
            out = sys.stdout if t[1] is None else open(t[1], 'w')
            print 'Reading', t[0], '...\n'
            input_data = read_data_function(t[0], out)
            run_function(input_data, out)
            out.close()


