'''
Common functions and objects for two simplex method algorithms
'''

# matrix library
import numpy as np
# linear algebra from matrix library
import numpy.linalg as alg
# python library for fractions
import fractions
from fractions import Fraction

class Delta:
    '''
    Stores information about calculated delta
    '''
    def __init__(self, index, value):
        '''
        Object constructor
            :param index: index of corresponding variable (x<index>)
            :param value: value of delta
        '''
        self.__index = index
        self.__value = value
    def __str__(self):
        '''
        String representation for str(obj) call
        '''
        return "(x%s, %s)" % (self.__index + 1, self.__value)
    def __repr__(self):
        '''
        String representation for default python string conversion
        '''
        return "(x%s, %s)" % (self.__index + 1, self.__value)
    def get_value(self):
        '''
        Returns value of delta
        '''
        return self.__value
    def get_index(self):
        '''
        Returns index of corresponding variable (x<index>)
        '''
        return self.__index   

# Two kinds of thetas - theta1 and theta2. Their differences are used by algorithm.
THETA_KIND_1 = 1
THETA_KIND_2 = 2

class Theta:
    '''
    Stores information about calculated theta
    '''
    def __init__(self, delta, table_index, index, divident, divider, kind):
        '''
        Object constructor
            :param delta: corresponding delta valeu (Delta object)
            :param table_index: index of corresponding variable in our table
            :param index: index of corresponding variable (x<index>)
            :param value: value of theta
            :param kind: kind of theta: theta1 or theta2 (see THETA_KIND_1 and THETA_KIND_2)
        '''
        self.__delta = delta
        self.__table_index = table_index
        if not (divider == 0 or divident == float("inf") or divident == float("-inf")):
            self.__value = Fraction(int(round(divident)), int(round(divider)))
        else:
            self.__value = divident / divider
        self.__index = index
        self.__kind = kind

    def __str__(self):
        '''
        String representation for str(obj) call
        '''
        return '(x%s, %s)' %  (self.__index + 1, self.__value, )

    def __repr__(self):
        '''
        String representation for default python string conversion
        '''
        return '(x%s, %s)' %  (self.__index + 1, self.__value, )

    def get_delta(self):
        '''
        Returns corresponding Delta value 
        '''
        return self.__delta
    def get_value(self):
        '''
        Returns calculated theta value
        '''
        return self.__value
    def get_index(self):
        '''
        Returns index of corresponding variable (x<index>)
        '''
        return self.__index
    def get_table_index(self):
        '''
        Returns of corresponding variable (x<index>) in our working table
        '''
        return self.__table_index
    def get_kind(self):
        '''
        Returns kind of theta (1 or 2), see THETA_KIND_1 and THETA_KIND_2
        '''
        return self.__kind
    def is_kind_1(self):
        '''
        Returns True if the theta is of kind 1 (see THETA_KIND_1), False otherwise
        '''
        return self.__kind == THETA_KIND_1
    def is_kind_2(self):
        '''
        Returns True if the theta is of kind 2 (see THETA_KIND_2), False otherwise
        '''
        return self.__kind == THETA_KIND_2

class WorkingEntities:
    '''
    Class for storing values needed for simplex method. It stores them as fields (without getters/setters) because 
    this class is used like 'storage' of different variables.
    '''
    def __init__(self, equations, target, bounds):        
        '''
        Creates instance of the class according to given input data.
            :param equations: main equations (for N variables and M equations - M lists of (N + 1) values)
            :param target: target function to be minimized (for N variables - list of N values)
            :param bounds: variable bounds (for N variables and M equations - list of N + M bounds)
        '''
        self.bounds = bounds 
        self.M = len(equations)
        self.N = len(equations[0]) - 1
        self.equations = np.matrix(equations)
        self.c = [int(t) for t in target]

        self.A = np.concatenate((np.matrix([t[:-1] for t in equations]), np.identity(self.M) * -1), axis=1)
        self.A_b = np.diag([-1 for i in range(0, self.M)])
        self.A_b_ = np.diag([1 for i in range(0, self.M)])
        self.A_b_0 = self.A_b_ * -1
        self.basis_indices = [i + self.N for i in range(0, self.M)]
        self.values = [0 for i in range(0, self.N + self.M)]
        self.x_b = np.matrix([0 for i in range(0, self.M)]).transpose()
        self.basis = np.matrix([0 for i in range(0, self.M + self.N)])
        self.u = np.matrix([-1 for i in range(0, self.M)], int)
        self.AbAj = np.matrix(np.zeros(self.M)).transpose()

    def get_bound(self, index):
        '''
        Returns bounds related to given variable.
            :param index: index of variable
        '''
        return self.bounds[index]

    def print_result(self, out):
        '''
        Prints result of calculation (values of each variable excluding synthetic + target function value)
            :param out: output stream to be used for printing
        '''
        print >> out, "Optimal solution was found:"
        for i in range(0, self.N):
            print >> out, 'x%s = %s' % (i + 1, self.values[i], )
        print >> out, "F =", multiply_vector_to_vector(self.c, self.values[0:self.N])

    def pretty_print(self, out, theta, f_value = False, fun_print = True):
        '''
        Prints current state of calculation as table
            :param out: output stream to be used for printing
            :param theta: selected theta
            :param f_value: True/False. Indicates if function value should be printed
        '''

        if not f_value:
            u_temp = self.u * self.A_b_0
        else:
            u_temp = self.u
        u_temp = u_temp.astype(int)
        delta_line = '' if theta is None else ('Delta(%d) = %s' % (theta.get_delta().get_index() + 1, theta.get_delta().get_value(), ))
        upper_line = ["", "" if not f_value else str(multiply_vector_to_vector([self.c[self.basis_indices[i]] for i in range(len(self.x_b))], [self.x_b[i, 0] for i in range(len(self.x_b))])) ] + [' ' if not fun_print else str(u_temp[0, val]) for val in range(0, u_temp.shape[1])] + [delta_line]
        middle_lines = list()
        for i in range(0, self.A_b_0.shape[0]):
            middle_lines.append(['x`' + str(self.basis_indices[i] + 1), str(self.x_b[i, 0])] + 
                [str(self.A_b_0[i, j]) for j in range(0, self.A_b_0.shape[1])] + 
                [' ' if theta is None else str(self.AbAj[i, 0])])
    
        all_lines = [upper_line] + middle_lines    
        gaps = list()
        for i in range(0, len(upper_line)):
            gaps.append(max([len(all_lines[k][i]) for k in range(0, len(all_lines))]))

    
        total = sum(gaps)
        divider_length = total + 3 * len(gaps) + 1
        print >> out,  '_' * divider_length +  '\n'
        for i in range(0, len(all_lines)):
            s = "|"
            for j in range(0, len(all_lines[i])):
                s = s + ((' {:' + str(gaps[j]) + 's} ').format(all_lines[i][j])) + '|'
            print >> out, s
            print >> out,  '_' * divider_length +  '\n'        

    def deltas_for_step_1(self):
        '''
        Calculates list of deltas (for step 1).
        '''
        estimation=list()
        temp = self.u * self.A_b_0
        for i in range(0, self.N):
            est = temp * self.extract_original_column(i)
            estimation.append(Delta(i, int(round(est.item(0)))))
        return estimation

    def deltas_for_step_2(self):
        '''
        Calculates list of deltas (for step 2)
        '''
        estimation=list()
        for i in range(0, self.N):
            if i in self.basis_indices: continue
            est = (self.u * self.extract_original_column(i))[0, 0] - int(round(alg.det(self.A_b))) * self.c[i]
            estimation.append(Delta(i, int(round(est.item(0)))))
        return estimation

    def filter_deltas(self, estimation):
        '''
        Removes deltas accoding to their bounds.
            :param estimation: list of calculated deltas
        '''
        available_for_improvement = list()
        for i in range(0, len(estimation)):
            bound = self.get_bound(estimation[i].get_index())
            if bound.isL() and estimation[i].get_value() < 0: continue #L
            if bound.isU() and estimation[i].get_value() > 0: continue #U
            if bound.isZ() and estimation[i].get_value() == 0: continue #Z
            available_for_improvement.append(estimation[i])
        return available_for_improvement

    def extract_original_column(self, i, t = int):
        '''
        Extracts column from original equation matrix.
            :param i: index of colum nto be extracted
            :param t: type of containing elements, int by default
        '''
        return self.equations[:,i].astype(t)

    def calculate_thetas(self, idx):
        '''
        Calculates thetas for given delta
            :param idx: corresponding delta
        '''
        detA_b = int(round(alg.det(self.A_b)))
        exp_AbAj = self.A_b_0 * self.extract_original_column(idx.get_index())
        theta_1 = list()
        theta_2 = list()
        for i in range(0, len(exp_AbAj)):
            cur = self.basis_indices[i]
            bound = self.bounds[cur]
            x_i = self.x_b.item(i)
            l_i = bound.get_lower_value()
            u_i = bound.get_upper_value()
            z_i = exp_AbAj[i, 0]
            if exp_AbAj[i]> 0 and idx.get_value() > 0:
                theta_1.append(Theta(idx, i, cur, (x_i - abs(detA_b) * l_i) , z_i, THETA_KIND_1))
            elif exp_AbAj[i] < 0 and idx.get_value() > 0:
                theta_2.append(Theta(idx, i, cur, (x_i - abs(detA_b) * u_i) , z_i, THETA_KIND_2))
            elif exp_AbAj[i]> 0 and idx.get_value() < 0:
                theta_1.append(Theta(idx, i, cur, (u_i - abs(detA_b) * x_i) , z_i, THETA_KIND_1))
            elif exp_AbAj[i]< 0 and idx.get_value() < 0:
                theta_2.append(Theta(idx, i, cur, (l_i - abs(detA_b) * x_i) , z_i, THETA_KIND_2))
        return (theta_1, theta_2)

    def recalculate_AbAj(self, index):
        '''
        Calculates new AbAj and stores this value
            :param index: selected index of variable (column to be used)
        '''
        self.AbAj = self.A_b_0 * self.extract_original_column(index)

#def divideIntOrFloat(val1, val2):
#    return val1 / val2 if val1 % val2 == 0 : 

def make_Q_transformation(out, A_b, column, rotation_index):
    '''
    Creates and executes Q-transformation.
       :param out: output stream to be used for printing
       :param A_b: A_b
       :param column: column to be used for replacing
       :param rotation_index: index for rotation (in terms of working table)
    '''
    print >> out, "Building Q-matrix..."
    Q_base = np.concatenate((A_b, np.identity(A_b.shape[0], int), column), axis=1)
    print >> out, Q_base

    print >> out, "..."
    Q0 = np.zeros(Q_base.shape, int)
    for i in range(0, Q0.shape[0]):
        for j in range(0, Q0.shape[1]):        
             Q0[i,j] = int(round(alg.det(replace_column(A_b, i, Q_base[:,j:j+1]))))
    print >> out, Q0
    print >> out, "Rotating column", rotation_index, "..."

    detB = alg.det(A_b)
    Q_rotated = np.zeros(Q0.shape, int)
    for i in range(0, Q_rotated.shape[0]):
        for j in range(0, Q_rotated.shape[1]):
            if i == rotation_index:
                Q_rotated[i,j] = int(round(Q0[i,j]))
            else:
                Q_rotated[i,j] = int(round((Q0[i, j] * Q0[rotation_index, Q0.shape[1]-1] - Q0[rotation_index, j] * Q0[i, Q0.shape[1]-1]) / detB))    

    print >> out, Q_rotated
    print >> out, "Next basis matrix determinant =", Q_rotated[rotation_index,Q_rotated.shape[1] - 1]
    
    return Q_rotated[:,[A_b.shape[0] + i for i in range(0, A_b.shape[0])]]


def update_removed_index_value(values, theta, bounds):
    '''
    Chooses new value for removed variable
        :param values: list of values to be updated
        :param theta: selected theta
        :param bounds: bounds for given variable
    '''
    delta_value = theta.get_delta().get_value()
    theta_index = theta.get_index()    
    if theta.is_kind_1() and delta_value < 0 :
        values[theta_index] = bounds.get_upper_value()
    elif theta.is_kind_1() and delta_value > 0 :
        values[theta_index] = bounds.get_lower_value()
    elif theta.is_kind_2() and delta_value < 0 :
        values[theta_index] = bounds.get_lower_value()
    elif theta.is_kind_2() and delta_value > 0 :
        values[theta_index] = bounds.get_upper_value()
    

def replace_column(matrix, i, j):
    '''
    Util method for temporary replacing of matrix column. Clones matrix and replaces column
        :param matrix: target matrix
        :param i: index of column to be replaced
        :param j: column to be used as replacement
    '''
    m = np.matrix(np.copy(matrix))
    m[:,i] = j
    return m


def multiply_vector_to_vector(v1, v2):
    '''
        Multiplies vectors as matrices
        :param v1: vector of size N
        :param v2: vector of size N
    '''
    result = 0
    for i in range(len(v1)):
        result += v1[i] * v2[i]
    return result

