'''
Input data parser
'''
# system library
import sys
# functional tools library
import functools

# set of constants for representing bounds (conditions)
LESS = 0
LESS_OR_EQ = 1
EQ = 2
MORE = 3
MORE_OR_EQ = 4

# text representation of bounds conditions (for printing)
boundary_type_representation = {LESS: "<", LESS_OR_EQ: "<=", EQ: "=", MORE: ">", MORE_OR_EQ: ">="}

class Bounds:    
    '''
    Represents common variable bounds (interval)
    '''
    def __init__(self, lo = (MORE_OR_EQ, 0.0), up = (LESS, float("+inf"))):
        '''
        Creates bound
            :param lo: lower bound, more or equals to zero by default
            :param up: upper bound, less then positive infinity by default
        '''
        self.l = lo
        self.u = up

    def __str__(self):
        '''
        String representation for str(obj) call
        '''
        return '%s %s X; %s %s X; L: %s, U: %s, F: %s, Z: %s' % (self.l[1], boundary_type_representation[self.l[0]], self.u[1], boundary_type_representation[self.u[0]], self.isL(), self.isU(), self.isF(), self.isZ(),)

    def get_lower_value(self):
        '''
        Returns lower bound value
        '''
        return self.l[1]

    def get_upper_value(self):
        '''
        Returns upper bound value
        '''
        return self.u[1]

    def isL(self):
        '''
        Returns true if this variable in L (lower value is infinity)
        '''
        return not self.l[1] == float("-inf") 
    def isU(self):
        '''
        Returns true if this variable in U (upper value is infinity)
        '''
        return not self.u[1] == float("inf") 
    def isF(self):
        '''
        Returns true if this variable in F (lower value equals to upper value)
        '''
        return self.l[1] == self.u[1]
    def isZ(self):
        '''
        Returns true if this variable in Z (lower value is negative infinity, upper is positive infinity)
        '''
        return self.l[1] == float("-inf") and self.u[1] == float("inf") 

class InputData:
    '''
    Represents problem's data
    '''
    def __init__(self, N, M):
        '''
        Creates problem's data with given variable and equation number
            :param N: variables number
            :param N: equations number
        '''
        self.__N = N
        self.__M = M
        self.__equations = list()
        self.__bounds = dict()
        self.__target = None

    def __str__(self):
        '''
        String representation for str(obj) call
        '''
        result = "%s variables %s equations %s" % (self.__N, self.__M, )
        result = result + 'Equations:\n'
        for t in self.__equations:
            result = result + '    ' + str(t) + '\n'
        result = result + 'Bounds:\n'
        for t in self.__bounds:
            result = result + '    ' + str(t) + ": " + str(self.__bounds[t]) + '\n'
        result = result + 'Target function:' + str(self.__target) + '\n'
        return result
    
    def append_equation(self, equation):
        '''
        Appends new equation. Equation is represented by N variable coefficients and value
            :param equation: given equation
        '''
        self.__equations.append(equation)

    def create_bounds(self):
        '''
        Creates default bounds (xi >= 0)
        '''
        for i in range(0, self.__N):
            self.__bounds[i] = Bounds()

    def add_bound(self, i, l, u):
        '''
        Sets bound for specific variable
            :param i: variable index
            :param l: lower bound in format (kind, value)
            :param u: upper bound in format (kind, value)
        '''
        self.__bounds[i] = Bounds(l, u)

    def add_bound_more_or_eq(self, i, val):
        '''
        Sets bound for specific variable: from given value (inclusive) to positive infinity
            :param i: variable index
            :param val: lower bound
        '''
        self.add_bound(i, (MORE_OR_EQ, val), (LESS, float("inf")))

    def add_bound_less_or_eq(self, i, val):
        '''
        Sets bound for specific variable: from negative infinity to given value (inclusive)
            :param i: variable index
            :param val: upper bound
        '''
        self.add_bound(i, (MORE, float("-inf")), (LESS_OR_EQ, val))

    def add_bound_open(self, i):
        '''
        Sets bound for specific variable: from negative infinity to positive infinity
            :param i: variable index
        '''
        self.add_bound(i, (MORE, float("-inf")), (LESS, float("inf")))

    def set_target(self, target):
        '''
        Sets target function. Function is represented by N variable coefficients.
            :param target: given function
        '''
        self.__target = list(target)

    def get_N(self):
        '''
        Returns count of variables
        '''
        return self.__N;

    def get_M(self):
        '''
        Returns count of equations
        '''
        return self.__M;

    def get_Equations(self):
        '''
        Returns copy of list of all equations
        '''
        return self.__equations[:];
    
    def get_Bounds(self):
        '''
        Returns present bounds
        '''
        return self.__bounds;

    def get_Target(self):
        '''
        Returns copy of list representing target function
        '''
        return self.__target[:];

def find_dividers(value):
    '''
    Finds all dividers of given integer number. Combinates them into dictionary of format 'divider' -> 'power'
        :param value: given value
    '''
    result = []
    while value > 1:
        found = False
        for i in range(2, value / 2 + 1):
            if value % i == 0:
                value = value / i
                result.append(i) 
                found = True
                break 
        if not found:
            result.append(value)
            break
    return dict([(el, result.count(el)) for el in set(result)])

def find_common_dividers(value1, value2):
    '''
    Finds common dividers for two numbers (int)
        :param value1: first value
        :param value2: second value
    '''
    d1 = find_dividers(value1)
    d2 = find_dividers(value2)
    common = set(t for t in d1) & set(t for t in d2)
    return [(t, min(d1.get(t, 0), d2.get(t, 0))) for t in common]

class Coefficient:
    '''
    Represents fraction coefficient
    '''
    def __init__(self, denominator, divider = 1):
        '''
        Creates fraction coefficient from denominator and divider. If they have common dividers, removes them.
            :param denominator: denominator (int)
            :param divider: divider(int), default value is 1
        '''
        sign = signum(denominator) * signum(divider)
        common_divider = functools.reduce(lambda x, y: x * (y[0] ** y[1]), find_common_dividers(denominator, divider), 1)
        self.__denominator = sign * abs(denominator) / common_divider;
        self.__divider = abs(divider) / common_divider;
    def __str__(self):
        '''
        String representation for str(obj) call
        '''
        return str(self.__denominator) if self.__divider == 1 else "%s / %s" % (self.__denominator, self.__divider)
    def __repr__(self):
        '''
        String representation for default python string conversion
        '''
        return str(self.__denominator) if self.__divider == 1 else "%s / %s" % (self.__denominator, self.__divider)
    def get_divider(self):
        '''
        Returns divider part
        '''
        return self.__divider
    def get_value(self):
        '''
        Returns value of coefficient as number
        '''
        return float(self.__denominator) / float(self.__divider)
    def multiply(self, val):
        '''
        Multiplies coefficient by given value, returns new coefficient
            :param val: multiplier
        '''
        return Coefficient(self.__denominator * val, self.__divider)
    def abs(self):
        '''
        Returns absolute value of the coefficient
        '''
        return Coefficient(abs(self.__denominator), abs(self.__divider))

def signum(val):
    '''
    Returns signum of given value
        :param val: given value
    '''
    return 0 if val == 0 else val / abs(val)

def parse_fraction(val):
    '''
    Parces fraction value form input
        :param val: string to be parsed as fraction
    '''
    if '/' in val:
        spl = val.split("/")
        return Coefficient(int(spl[0]), int(spl[1]))
    else:
        return Coefficient(int(val))

def coefficient_functions_as_str(equation):
    '''
    Represents set of coefficients as equation string
        :param equation: coefficients for corresponding variables
    '''
    result = ""
    for i in range(0, len(equation)):
        val = equation[i] if type(equation[i]) is int else equation[i].get_value()
        abs_val = abs(equation[i]) if type(equation[i]) is int else equation[i].abs()
        if result == "":
            if not val == 0:
                result = result + "%s * x%s" % (str(equation[i]), i, )
        else:
            if val > 0:
                result = result + " + %s * x%s" % (str(abs_val), i, )
            elif val < 0:
                result = result + " - %s * x%s" % (str(abs_val), i, )
    return result

def equation_as_str(equation):
    '''
    Represents equation as string
        :param equation: equation to be represented
    '''
    return coefficient_functions_as_str(equation[:-1]) + " = " + str(equation[-1])

def target_as_str(equation):
    '''
    Represents target function as string
        :param equation: function coeffitients to be represented
    '''
    return coefficient_functions_as_str(equation) + " -> min"

def transform(out, equation_as_str_fun, equation):
    '''
    Transforms equation with fractional coefficients into integer-coefficients equation
        :param out: output stream for printing
        :param equation: equation to be transformed
    '''
    if all(t.get_divider() == 1 for t in equation): return [t.get_value() for t in equation]
    dividers = [t.get_divider() for t in equation]
    multiplier = dict()
    for value in dividers:
        d = find_dividers(value)
        for i in d:
            multiplier[i] = max(multiplier.get(i, 0), d[i])
    multiplier = functools.reduce(lambda x, y: x * (y ** multiplier[y]), multiplier, 1)
    result = [int(t.multiply(multiplier).get_value()) for t in equation]
    print >> out, "Transforming:" 
    print >> out, equation_as_str_fun(equation), "to"
    print >> out, equation_as_str_fun(result)
    return result

def read_input_data(fname, out):
    '''
    Reads data from given file: target function and equations
        :param fname: name of file
    '''
    with open(fname) as f:
        content = f.readlines()
    
    # removes empty lines at the end of file
    last_empty_line = None
    for i in range(0, len(content)):
        if content[len(content) - 1 - i].strip() == "":
            last_empty_line = len(content) - i - 1
        else:
            break
    if last_empty_line is not None:
        content = content[:last_empty_line]

    # detect number of variables and equations
    first_line = content[0].split()
    N = len(first_line)
    M = len(content) - 1
    input_data = InputData(N, M)
    
    # parses target function (first line)
    input_data.set_target(transform(out, target_as_str, [parse_fraction(t) for t in first_line]))

    # parses equations
    for i in range(0, M):
        spl = content[i + 1].split()
        equation = transform(out, equation_as_str, [parse_fraction(t) for t in spl])
        if not len(equation) == N + 1:
            sys.exit('Error at line %s: expected %s elements, found %s. Equation should consists of %s coefficients and 1 value' % (i + 1, N + 1, len(equation), N,))
        input_data.append_equation(equation);


    return input_data
    
def read_linear_system(fname, out):
    '''
    Reads data from given file:  N equations with N vars
        :param fname: name of file
    '''
    with open(fname) as f:
        content = f.readlines()
    
    # removes empty lines at the end of file
    last_empty_line = None
    for i in range(0, len(content)):
        if content[len(content) - 1 - i].strip() == "":
            last_empty_line = len(content) - i - 1
        else:
            break
    if last_empty_line is not None:
        content = content[:last_empty_line]

    # detect number of variables and equations
    first_line = content[0].split()
    N = len(first_line)
    if not N == len(content) + 1:
        sys.exit('Error: file contains worng data: %s equations and %s variables.' % (len(content), N - 1))
    input_data = list()
    
    # parses equations
    for i in range(0, N - 1):
        spl = content[i].split()
        equation = transform(out, equation_as_str, [parse_fraction(t) for t in spl])
        if not len(equation) == N:
            sys.exit('Error at line %s: expected %s elements, found %s. Equations should have the same number of elements.' % (i, N, len(equation),))
        input_data.append(equation);


    return input_data
    

