# matrix library
import numpy as np
# linear algebra from matrix library
import numpy.linalg as alg
# library for object reflection
import operator
# common functions
import common
from common import Delta, Theta, WorkingEntities, update_removed_index_value, make_Q_transformation, replace_column
# python library for fractions
import fractions
from fractions import Fraction

# List of return statuses of calculation steps
# continue (everything is ok)
STEP_RESULT_CONTINUE = 0
# end (end of phase)
STEP_RESULT_END = 1
# fail (no solution)
STEP_RESULT_FAIL = 2


def simplex_step_of_phase_1(out, entities):
        '''
        Executes single step of phase 1
            :param out: output stream to be used for printing
            :param entities: working entities

            :return int: status of executed step, see STEP_RESULT_CONTINUE, STEP_RESULT_END, STEP_RESULT_FAIL
        '''
        estimation = entities.deltas_for_step_1()
        print >> out, "Delta:", estimation
        available_for_improvement = entities.filter_deltas(estimation)
        if len(available_for_improvement) == 0:
            return STEP_RESULT_END
   
        max_estimation_index = max(available_for_improvement,key=operator.methodcaller('get_value'))
        print >> out, "Chosen:", "x" + str(max_estimation_index.get_index() + 1), " with delta", max_estimation_index.get_value()

        theta_1, theta_2 = entities.calculate_thetas(max_estimation_index)
        print >> out, "Theta:", theta_1, "and", theta_2

        if len(theta_1 + theta_2) == 0:
            return STEP_RESULT_END
        elif all(i.get_value() == float('inf') for i in (theta_1 + theta_2)):
            print >> out, "No non-infinitive theta was found."
            entities.pretty_print(out, None)
            print >> out, "No solution have been found: objective function is unbounded."
            return STEP_RESULT_FAIL

        theta = min(theta_1 + theta_2,key=operator.methodcaller('get_value'))
        print >> out, "Chosen theta:", theta.get_value(), "for", "x" + str(theta.get_index() + 1)
        entities.recalculate_AbAj(max_estimation_index.get_index())

        x_to_remove = theta.get_index()
        x_to_add = max_estimation_index.get_index()
        print >> out, "Replacing", "x" + str(x_to_remove + 1), " with ", "x" + str(x_to_add + 1)

        entities.pretty_print(out, theta)

        column_to_add = entities.extract_original_column(x_to_add)
        # calculate new matrices
        new_A_b_ = make_Q_transformation(out, entities.A_b, column_to_add, theta.get_table_index())
        new_A_b = np.copy(entities.A_b)
        new_A_b[:,theta.get_table_index():theta.get_table_index() + 1] = column_to_add

        new_values = entities.values[:]
        new_basis_indices = entities.basis_indices[:]

        update_removed_index_value(new_values, theta, entities.get_bound(theta.get_index()))
        new_basis_indices[theta.get_table_index()] = theta.get_delta().get_index()
        non_basis_values = list()
        for i in range(0, len(new_values)):
            if i not in new_basis_indices:
                if i >= entities.N:
                    non_basis_values.append(new_values[i])
                else:
                    non_basis_values.append(new_values[i])
            else:
                non_basis_values.append(0)
        non_basis_values = np.matrix(non_basis_values).transpose()
        x_b_ = - int(np.sign(alg.det(new_A_b))) * new_A_b_ * (entities.A * non_basis_values).astype(int)
        for i in range(0, x_b_.shape[0]):
            new_values[new_basis_indices[i]] = Fraction(x_b_[i, 0], abs(int(round(alg.det(new_A_b)))))
        new_x_b = x_b_
        new_A_b_0 = int(np.sign(alg.det(new_A_b))) * new_A_b_
        new_u = np.copy(entities.u)
        new_u[0, theta.get_table_index()] = 0

        #update working entities for next step
        (entities.A_b, entities.A_b_, entities.A_b_0, entities.u, entities.values, entities.x_b, entities.basis_indices) = (new_A_b, new_A_b_, new_A_b_0, np.matrix(new_u), new_values, new_x_b, new_basis_indices)
        return STEP_RESULT_CONTINUE



def simplex_phase_1(out, entities):
    '''
    Executes phase 1
        :param out: output stream to be used for printing
        :param entities: working entities

        :return True/False: True in the case of success (continue with next phase), False in the case of no solution
    '''
    iteration = 1
    proceed = STEP_RESULT_CONTINUE
    while proceed == STEP_RESULT_CONTINUE:
        print >> out, "=== Phase 1", "Iteration", iteration, ".\n"
        proceed = simplex_step_of_phase_1(out, entities)
        iteration = iteration + 1

    for var in entities.basis_indices:
        if var >= entities.N:
            print >> out, "Synthetic variable is present:", "x" + str(var + 1)
            if entities.AbAj[entities.basis_indices.index(var), 0] == 0:
                print >> out, "Linear system is inconsistent."
                return False
            else:
                print >> out, "Linear system is inconsistent."
                return False

    return True


def simplex_step_of_phase_2(out, entities):
        '''
        Executes single step of phase 2
            :param out: output stream to be used for printing
            :param entities: working entities

            :return int: status of executed step, see STEP_RESULT_CONTINUE, STEP_RESULT_END, STEP_RESULT_FAIL
        '''
        entities.u = np.matrix([entities.c[i] for i in entities.basis_indices]) * entities.A_b_0
        estimation = entities.deltas_for_step_2()
        print >> out, "Delta:", estimation
        available_for_improvement = entities.filter_deltas(estimation)
        if len(available_for_improvement) == 0:
            print >> out, "No improvements are available.\n"
            entities.pretty_print(out, None, True)
            entities.print_result(out)
            return STEP_RESULT_END
   
        max_estimation_index = max(available_for_improvement,key=operator.methodcaller('get_value'))
        print >> out, "Chosen:", "x" + str(max_estimation_index.get_index() + 1), " with delta", max_estimation_index.get_value()

        theta_1, theta_2 = entities.calculate_thetas(max_estimation_index)
        print >> out, "Theta:", theta_1, "and", theta_2

        if len(theta_1 + theta_2) == 0:
            entities.print_result(out)
            return STEP_RESULT_END
        elif all(i.get_value() == float('inf') for i in (theta_1 + theta_2)):
            print >> out, "No non-infinitive theta was found."
            entities.pretty_print(out, None, True)
            print >> out, "No solution have been found: objective function is unbounded."
            return STEP_RESULT_FAIL

        theta = min(theta_1 + theta_2,key=operator.methodcaller('get_value'))
        print >> out, "Chosen theta:", theta.get_value(), "for", "x" + str(theta.get_index() + 1)
        entities.recalculate_AbAj(max_estimation_index.get_index())

        x_to_remove = theta.get_index()
        x_to_add = max_estimation_index.get_index()
        print >> out, "Replacing", "x" + str(x_to_remove + 1), " with ", "x" + str(x_to_add + 1)

        entities.pretty_print(out, theta, True)

        column_to_add = entities.extract_original_column(x_to_add)
        # calculate new matrices
        new_A_b_ = make_Q_transformation(out, entities.A_b, column_to_add, theta.get_table_index())
        new_A_b = np.copy(entities.A_b)
        new_A_b[:,theta.get_table_index():theta.get_table_index() + 1] = column_to_add

        new_values = entities.values[:]
        new_basis_indices = entities.basis_indices[:]

        update_removed_index_value(new_values, theta, entities.get_bound(theta.get_index()))
        new_basis_indices[theta.get_table_index()] = theta.get_delta().get_index()
        non_basis_values = list()
        for i in range(0, len(new_values)):
            if i not in new_basis_indices:
                if i >= entities.N:
                    non_basis_values.append(new_values[i])
                else:
                    non_basis_values.append(new_values[i])
            else:
                non_basis_values.append(0)
        non_basis_values = np.matrix(non_basis_values).transpose()
        x_b_ = - int(np.sign(alg.det(new_A_b))) * new_A_b_ * (entities.A * non_basis_values).astype(int)
        for i in range(0, x_b_.shape[0]):
            new_values[new_basis_indices[i]] = Fraction(x_b_[i, 0], abs(int(round(alg.det(new_A_b)))))
        new_x_b = x_b_
        new_A_b_0 = int(np.sign(alg.det(new_A_b))) * new_A_b_

        #update working entities for next step
        (entities.A_b, entities.A_b_, entities.A_b_0, entities.values, entities.basis_indices, entities.x_b)    = (new_A_b, new_A_b_, new_A_b_0, new_values, new_basis_indices, new_x_b)
        return STEP_RESULT_CONTINUE


def simplex_phase_2(out, entities):
    '''
    Executes phase 2
        :param out: output stream to be used for printing
        :param entities: working entities

        :return True/False: True in the case of success, False in the case of no solution
    '''
    iteration = 1
    proceed = STEP_RESULT_CONTINUE
    while proceed == STEP_RESULT_CONTINUE:
        print >> out, "=== Phase 2", "Iteration", iteration, ".\n"
        proceed = simplex_step_of_phase_2(out, entities)
        iteration = iteration + 1


def run_simplex(input_data, out):
    '''
    Runs simplex method
        :param input_data: parsed input data
        :param out: output stream (for log and solution printing)
    '''

    for i in range(0, input_data.get_N()):
        input_data.add_bound_more_or_eq(i, 0)
    for i in range(0, input_data.get_M()):
        input_data.add_bound_less_or_eq(input_data.get_N() + i, input_data.get_Equations()[i][-1])

    entities = WorkingEntities(input_data.get_Equations(), input_data.get_Target(), input_data.get_Bounds())
    success = simplex_phase_1(out, entities)
    if success:
        simplex_phase_2(out, entities)



	
